# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-06-26 04:31
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import versionfield
import resolwe.flow.models.fields


class Migration(migrations.Migration):

    replaces = [
        ('flow', '0001_initial'),
        ('flow', '0002_project_to_collection'),
        ('flow', '0003_support_sample'),
        ('flow', '0004_autoslug_field'),
        ('flow', '0005_process_data_name'),
        ('flow', '0006_data_named_by_user'),
        ('flow', '0007_add_owner'),
        ('flow', '0008_fix_jsonfields'),
        ('flow', '0009_data_parents'),
        ('flow', '0010_fix_jsonfields'),
        ('flow', '0011_calculate_checksum'),
        ('flow', '0012_require_checksum'),
        ('flow', '0013_add_requirements'),
        ('flow', '0014_add_entity'),
        ('flow', '0015_make_data_indexes'),
        ('flow', '0016_update_versionfield'),
        ('flow', '0017_update_checksum'),
        ('flow', '0018_remove_triggers'),
        ('flow', '0019_data_descriptor_dirty'),
        ('flow', '0020_collection_descriptor_dirty'),
        ('flow', '0021_entity_descriptor_completed'),
        ('flow', '0022_data_sha1_to_sha256'),
        ('flow', '0023_update_checksum'),
        ('flow', '0024_add_relations'),
        ('flow', '0025_set_get_last_by'),
        ('flow', '0026_tags'),
        ('flow', '0027_scheduling_class'),
        ('flow', '0028_remove_public_processes'),
        ('flow', '0029_data_checksum_index'),
        ('flow', '0030_change_slug_field')
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Collection',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('description', models.TextField(blank=True)),
                ('settings', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor_dirty', models.BooleanField(default=False)),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('view_collection', 'Can view collection'), ('edit_collection', 'Can edit collection'), ('share_collection', 'Can share collection'), ('download_collection', 'Can download files from collection'), ('add_collection', 'Can add data objects to collection'), ('owner_collection', 'Is owner of the collection')),
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Data',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('started', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('finished', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('checksum', models.CharField(db_index=True, max_length=64, validators=[django.core.validators.RegexValidator(code='invalid_checksum', message='Checksum is exactly 40 alphanumerics', regex='^[0-9a-f]{64}$')])),
                ('status', models.CharField(choices=[('UP', 'Uploading'), ('RE', 'Resolving'), ('WT', 'Waiting'), ('PR', 'Processing'), ('OK', 'Done'), ('ER', 'Error'), ('DR', 'Dirty')], default='RE', max_length=2)),
                ('process_pid', models.PositiveIntegerField(blank=True, null=True)),
                ('process_progress', models.PositiveSmallIntegerField(default=0)),
                ('process_rc', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('process_info', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=[], size=None)),
                ('process_warning', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=[], size=None)),
                ('process_error', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=[], size=None)),
                ('input', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('output', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor_dirty', models.BooleanField(default=False)),
                ('named_by_user', models.BooleanField(default=False)),
                ('tags', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=list, size=None)),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('view_data', 'Can view data'), ('edit_data', 'Can edit data'), ('share_data', 'Can share data'), ('download_data', 'Can download files from data'), ('owner_data', 'Is owner of the data')),
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='DescriptorSchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('description', models.TextField(blank=True)),
                ('schema', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('view_descriptorschema', 'Can view descriptor schema'), ('edit_descriptorschema', 'Can edit descriptor schema'), ('share_descriptorschema', 'Can share descriptor schema'), ('owner_descriptorschema', 'Is owner of the description schema')),
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Entity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('description', models.TextField(blank=True)),
                ('settings', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor_dirty', models.BooleanField(default=False)),
                ('descriptor_completed', models.BooleanField(default=False)),
                ('tags', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=list, size=None)),
                ('collections', models.ManyToManyField(to='flow.Collection')),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('data', models.ManyToManyField(to='flow.Data')),
                ('descriptor_schema', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema')),
            ],
            options={
                'permissions': (('view_entity', 'Can view entity'), ('edit_entity', 'Can edit entity'), ('share_entity', 'Can share entity'), ('download_entity', 'Can download files from entity'), ('add_entity', 'Can add data objects to entity'), ('owner_entity', 'Is owner of the entity')),
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='PositionInRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Entity')),
            ],
        ),
        migrations.CreateModel(
            name='Process',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('type', models.CharField(max_length=100, validators=[django.core.validators.RegexValidator(code='invalid_type', message='Type may be alphanumerics separated by colon', regex='^data:[a-z0-9:]+:$')])),
                ('category', models.CharField(default='other', max_length=200, validators=[django.core.validators.RegexValidator(code='invalid_category', message='Category may be alphanumerics separated by colon', regex='^([a-z0-9]+[:\\-])*[a-z0-9]+:$')])),
                ('persistence', models.CharField(choices=[('RAW', 'Raw'), ('CAC', 'Cached'), ('TMP', 'Temp')], default='RAW', max_length=3)),
                ('description', models.TextField(default='')),
                ('data_name', models.CharField(blank=True, max_length=200, null=True)),
                ('input_schema', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=list)),
                ('output_schema', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=list)),
                ('flow_collection', models.CharField(blank=True, max_length=100, null=True)),
                ('run', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('requirements', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('scheduling_class', models.CharField(choices=[('IN', 'Interactive'), ('BA', 'Batch')], default='BA', max_length=2)),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('view_process', 'Can view process'), ('share_process', 'Can share process'), ('owner_process', 'Is owner of the process')),
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Relation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('label', models.CharField(blank=True, max_length=100, null=True)),
                ('collection', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='flow.Collection')),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('entities', models.ManyToManyField(through='flow.PositionInRelation', to='flow.Entity')),
            ],
            options={
                'permissions': (('view_relation', 'Can view relation'), ('edit_relation', 'Can edit relation'), ('share_relation', 'Can share relation'), ('owner_relation', 'Is owner of the relation')),
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='RelationType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
                ('ordered', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='Storage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', resolwe.flow.models.fields.ResolweSlugField(max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('json', django.contrib.postgres.fields.jsonb.JSONField()),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('data', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Data')),
            ],
            options={
                'get_latest_by': 'version',
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.AddField(
            model_name='relation',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='flow.RelationType'),
        ),
        migrations.AddField(
            model_name='positioninrelation',
            name='relation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Relation'),
        ),
        migrations.AddField(
            model_name='data',
            name='descriptor_schema',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema'),
        ),
        migrations.AddField(
            model_name='data',
            name='parents',
            field=models.ManyToManyField(related_name='children', to='flow.Data'),
        ),
        migrations.AddField(
            model_name='data',
            name='process',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='flow.Process'),
        ),
        migrations.AddField(
            model_name='collection',
            name='data',
            field=models.ManyToManyField(to='flow.Data'),
        ),
        migrations.AddField(
            model_name='collection',
            name='descriptor_schema',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema'),
        ),
        migrations.AlterUniqueTogether(
            name='storage',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='relation',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='process',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='entity',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='descriptorschema',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='data',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='collection',
            unique_together=set([('slug', 'version')]),
        ),
    ]



"""SECCOMP policy.

This policy is based on the default Docker SECCOMP policy profile. It
allows several syscalls, which are most commonly used. We make one major
change regarding the network-related ``socket`` syscall in that we only
allow AF_INET/AF_INET6 SOCK_DGRAM/SOCK_STREAM sockets for TCP and UDP
protocols.
"""
# pylint: disable=too-many-lines
SECCOMP_POLICY = {
    "defaultAction": "SCMP_ACT_ERRNO",
    "architectures": [
        "SCMP_ARCH_X86_64",
        "SCMP_ARCH_X86",
        "SCMP_ARCH_X32"
    ],
    "syscalls": [
        {
            "name": "accept",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "accept4",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "access",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "alarm",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "bind",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "brk",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "capget",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "capset",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "chdir",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "chmod",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "chown",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "chown32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "clock_getres",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "clock_gettime",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "clock_nanosleep",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "close",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "connect",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "copy_file_range",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "creat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "dup",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "dup2",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "dup3",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_create",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_create1",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_ctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_ctl_old",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_pwait",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_wait",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "epoll_wait_old",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "eventfd",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "eventfd2",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "execve",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "execveat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "exit",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "exit_group",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "faccessat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fadvise64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fadvise64_64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fallocate",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fanotify_mark",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fchdir",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fchmod",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fchmodat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fchown",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fchown32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fchownat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fcntl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fcntl64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fdatasync",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fgetxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "flistxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "flock",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fork",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fremovexattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fsetxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fstat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fstat64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fstatat64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fstatfs",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fstatfs64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "fsync",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ftruncate",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ftruncate64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "futex",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "futimesat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getcpu",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getcwd",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getdents",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getdents64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getegid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getegid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "geteuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "geteuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getgid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getgroups",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getgroups32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getitimer",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getpeername",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getpgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getpgrp",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getpid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getppid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getpriority",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getrandom",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getresgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getresgid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getresuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getresuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getrlimit",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "get_robust_list",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getrusage",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getsid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getsockname",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getsockopt",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "get_thread_area",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "gettid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "gettimeofday",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "getxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "inotify_add_watch",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "inotify_init",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "inotify_init1",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "inotify_rm_watch",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "io_cancel",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ioctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "io_destroy",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "io_getevents",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ioprio_get",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ioprio_set",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "io_setup",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "io_submit",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ipc",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "kill",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lchown",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lchown32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lgetxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "link",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "linkat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "listen",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "listxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "llistxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "_llseek",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lremovexattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lseek",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lsetxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lstat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "lstat64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "madvise",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "memfd_create",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mincore",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mkdir",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mkdirat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mknod",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mknodat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mlock",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mlock2",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mlockall",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mmap",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mmap2",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mprotect",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mq_getsetattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mq_notify",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mq_open",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mq_timedreceive",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mq_timedsend",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mq_unlink",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "mremap",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "msgctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "msgget",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "msgrcv",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "msgsnd",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "msync",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "munlock",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "munlockall",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "munmap",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "nanosleep",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "newfstatat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "_newselect",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "open",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "openat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "pause",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "personality",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 0,
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                }
            ]
        },
        {
            "name": "personality",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 8,
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                }
            ]
        },
        {
            "name": "personality",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 4294967295,
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                }
            ]
        },
        {
            "name": "pipe",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "pipe2",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "poll",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ppoll",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "prctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "pread64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "preadv",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "prlimit64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "pselect6",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "pwrite64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "pwritev",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "read",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "readahead",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "readlink",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "readlinkat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "readv",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "recv",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "recvfrom",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "recvmmsg",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "recvmsg",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "remap_file_pages",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "removexattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rename",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "renameat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "renameat2",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "restart_syscall",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rmdir",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigaction",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigpending",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigprocmask",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigqueueinfo",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigreturn",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigsuspend",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_sigtimedwait",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "rt_tgsigqueueinfo",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_getaffinity",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_getattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_getparam",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_get_priority_max",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_get_priority_min",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_getscheduler",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_rr_get_interval",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_setaffinity",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_setattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_setparam",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_setscheduler",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sched_yield",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "seccomp",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "select",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "semctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "semget",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "semop",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "semtimedop",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "send",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sendfile",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sendfile64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sendmmsg",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sendmsg",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sendto",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setfsgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setfsgid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setfsuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setfsuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setgid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setgroups",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setgroups32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setitimer",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setpgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setpriority",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setregid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setregid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setresgid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setresgid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setresuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setresuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setreuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setreuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setrlimit",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "set_robust_list",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setsid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setsockopt",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "set_thread_area",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "set_tid_address",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setuid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setuid32",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "setxattr",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "shmat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "shmctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "shmdt",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "shmget",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "shutdown",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sigaltstack",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "signalfd",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "signalfd4",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sigreturn",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "socketpair",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "splice",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "stat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "stat64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "statfs",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "statfs64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "symlink",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "symlinkat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sync",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sync_file_range",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "syncfs",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "sysinfo",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "syslog",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "tee",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "tgkill",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "time",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timer_create",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timer_delete",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timerfd_create",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timerfd_gettime",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timerfd_settime",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timer_getoverrun",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timer_gettime",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "timer_settime",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "times",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "tkill",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "truncate",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "truncate64",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "ugetrlimit",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "umask",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "uname",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "unlink",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "unlinkat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "utime",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "utimensat",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "utimes",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "vfork",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "vmsplice",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "wait4",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "waitid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "waitpid",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "write",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "writev",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "arch_prctl",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "modify_ldt",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "chroot",
            "action": "SCMP_ACT_ALLOW",
            "args": []
        },
        {
            "name": "clone",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 2080505856,
                    "valueTwo": 0,
                    "op": "SCMP_CMP_MASKED_EQ"
                }
            ]
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 12,  # AF_INET
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 1,
                    "value": 1,  # SOCK_STREAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 2,
                    "value": 6,  # IPPROTO_TCP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
            ]
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 2,  # AF_INET
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 1,
                    "value": 2,  # SOCK_DGRAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 2,
                    "value": 17,  # IPPROTO_UDP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
            ]
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 10,  # AF_INET6
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 1,
                    "value": 1,  # SOCK_STREAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 2,
                    "value": 6,  # IPPROTO_TCP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
            ]
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 10,  # AF_INET6
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 1,
                    "value": 2,  # SOCK_DGRAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
                {
                    "index": 2,
                    "value": 17,  # IPPROTO_UDP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ"
                },
            ]
        },
    ]
}



""".. Ignore pydocstyle D400.

========
Consumer
========

Manager Channels consumer.

"""
import asyncio

import async_timeout
from channels.consumer import AsyncConsumer
from channels.layers import get_channel_layer
from channels.testing import ApplicationCommunicator

from . import state


async def send_event(message):
    """Construct a Channels event packet with the given message.

    :param message: The message to send to the manager workers.
    """
    packet = {
        'type': 'control_event',  # This is used as the method name in the consumer.
        'content': message,
    }
    await get_channel_layer().send(state.MANAGER_CONTROL_CHANNEL, packet)


async def run_consumer(timeout=None):
    """Run the consumer until it finishes processing."""
    channel = state.MANAGER_CONTROL_CHANNEL
    scope = {
        'type': 'control_event',
        'channel': channel,
    }
    from . import manager

    app = ApplicationCommunicator(ManagerConsumer, scope)

    channel_layer = get_channel_layer()

    async def _consume_loop():
        """Run a loop to consume messages off the channels layer."""
        while True:
            message = await channel_layer.receive(channel)
            if message.get('type', {}) == '_resolwe_manager_quit':
                break
            message.update(scope)
            await manager.sync_counter.inc('message-handling')
            await app.send_input(message)

    if timeout is None:
        await _consume_loop()
    try:
        # A further grace period to catch late messages.
        async with async_timeout.timeout(timeout or 1):
            await _consume_loop()
    except asyncio.TimeoutError:
        pass

    await app.wait()
    flush = getattr(channel_layer, 'flush', None)
    if flush:
        await flush()


async def exit_consumer():
    """Cause the synchronous consumer to exit cleanly."""
    packet = {
        'type': '_resolwe_manager_quit',
    }
    await get_channel_layer().send(state.MANAGER_CONTROL_CHANNEL, packet)


class ManagerConsumer(AsyncConsumer):
    """Channels consumer for handling manager events."""

    def __init__(self, *args, **kwargs):
        """Initialize a consumer instance with the given manager."""
        # This import is local in order to avoid startup import loops.
        from . import manager
        self.manager = manager
        super().__init__(*args, **kwargs)

    async def control_event(self, message):
        """Forward control events to the manager dispatcher."""
        try:
            await self.manager.handle_control_event(message['content'])
        finally:
            await self.manager.sync_counter.dec('message-handling')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-01-31 07:23
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0017_update_checksum'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='trigger',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='trigger',
            name='collection',
        ),
        migrations.RemoveField(
            model_name='trigger',
            name='contributor',
        ),
        migrations.RemoveField(
            model_name='trigger',
            name='process',
        ),
        migrations.DeleteModel(
            name='Trigger',
        ),
    ]



"""Elastic Search indexes for Data model."""
import elasticsearch_dsl as dsl

from resolwe.elastic.fields import Name, ProcessType
from resolwe.elastic.indices import BaseIndex

from ..models import Data
from .base import BaseDocument, BaseIndexMixin


class DataDocument(BaseDocument):
    """Document for data search."""

    started = dsl.Date()
    finished = dsl.Date()
    status = dsl.Keyword()
    process = dsl.Integer()
    process_type = ProcessType()
    # Keep backward compatibility.
    type = ProcessType()  # pylint: disable=invalid-name
    process_name = Name()
    tags = dsl.Keyword(multi=True)

    collection = dsl.Integer(multi=True)
    parents = dsl.Integer(multi=True)
    children = dsl.Integer(multi=True)
    entity = dsl.Integer(multi=True)

    class Meta:
        """Meta class for data search document."""

        index = 'data'


class DataIndex(BaseIndexMixin, BaseIndex):
    """Index for data objects used in ``DataDocument``."""

    queryset = Data.objects.all().prefetch_related(
        'process',
        'contributor'
    )
    object_type = Data
    document_class = DataDocument

    mapping = {
        'process': 'process.id',
        'process_name': 'process.name',
        'process_type': 'process.type',
        'type': 'process.type',
    }

    def get_dependencies(self):
        """Return dependencies, which should trigger updates of this model."""
        # pylint: disable=no-member
        return super().get_dependencies() + [
            Data.collection_set,
            Data.entity_set,
            Data.parents,
            Data.children,
        ]

    def get_collection_value(self, obj):
        """Extract contributor."""
        return list(obj.collection_set.values_list('pk', flat=True))

    def get_parents_value(self, obj):
        """Extract parents."""
        return list(obj.parents.values_list('pk', flat=True))

    def get_children_value(self, obj):
        """Extract children."""
        return list(obj.children.values_list('pk', flat=True))

    def get_entity_value(self, obj):
        """Extract entities."""
        return list(obj.entity_set.values_list('pk', flat=True))



""".. Ignore pydocstyle D400.

===========
Flow Models
===========

Base Model
==========

Base model for all other models.

.. autoclass:: resolwe.flow.models.base.BaseModel
    :members:

Collection Model
================

Postgres ORM model for the organization of collections.

.. autoclass:: resolwe.flow.models.collection.BaseCollection
    :members:

.. autoclass:: resolwe.flow.models.Collection
    :members:

Data model
==========

Postgres ORM model for keeping the data structured.

.. autoclass:: resolwe.flow.models.Data
    :members:

.. autoclass:: resolwe.flow.models.DataDependency
    :members:

DescriptorSchema model
======================

Postgres ORM model for storing descriptors.

.. autoclass:: resolwe.flow.models.DescriptorSchema
    :members:

Process model
=============

Postgres ORM model for storing processes.

.. autoclass:: resolwe.flow.models.Process
    :members:

Storage model
=============

Postgres ORM model for storing JSON.

.. autoclass:: resolwe.flow.models.Storage
    :members:

Secret model
============

Postgres ORM model for storing secrets.

.. autoclass:: resolwe.flow.models.Secret

ProcessMigrationHistory model
=============================

Postgres ORM model for storing proces migration history.

.. autoclass:: resolwe.flow.models.ProcessMigrationHistory
    :members:

DataMigrationHistory model
==========================

Postgres ORM model for storing data migration history.

.. autoclass:: resolwe.flow.models.DataMigrationHistory
    :members:

"""

from .collection import Collection
from .data import Data, DataDependency
from .descriptor import DescriptorSchema
from .entity import Entity, Relation, RelationType
from .migrations import DataMigrationHistory, ProcessMigrationHistory
from .process import Process
from .secret import Secret
from .storage import Storage



"""Descriptor schema viewset."""
from rest_framework import mixins, viewsets

from resolwe.flow.filters import DescriptorSchemaFilter
from resolwe.flow.models import DescriptorSchema
from resolwe.flow.serializers import DescriptorSchemaSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin


class DescriptorSchemaViewSet(mixins.RetrieveModelMixin,
                              mixins.ListModelMixin,
                              ResolwePermissionsMixin,
                              viewsets.GenericViewSet):
    """API view for :class:`DescriptorSchema` objects."""

    queryset = DescriptorSchema.objects.all().prefetch_related('contributor')
    serializer_class = DescriptorSchemaSerializer
    permission_classes = (get_permissions_class(),)
    filter_class = DescriptorSchemaFilter
    ordering_fields = ('id', 'created', 'modified', 'name', 'version')
    ordering = ('id',)



"""Reslowe process model."""
import copy
import json
import os

from django.conf import settings
from django.contrib.postgres.fields import ArrayField, JSONField
from django.core.exceptions import PermissionDenied, ValidationError
from django.core.validators import RegexValidator
from django.db import models, transaction

from resolwe.flow.expression_engines.exceptions import EvaluationError
from resolwe.flow.utils import dict_dot, get_data_checksum, iterate_fields, iterate_schema
from resolwe.permissions.utils import assign_contributor_permissions, copy_permissions

from .base import BaseModel
from .descriptor import DescriptorSchema
from .entity import Entity
from .secret import Secret
from .storage import Storage
from .utils import (
    DirtyError, hydrate_input_references, hydrate_size, render_descriptor, render_template, validate_schema,
)

# Compatibilty for Python < 3.5.
if not hasattr(json, 'JSONDecodeError'):
    json.JSONDecodeError = ValueError


class DataQuerySet(models.QuerySet):
    """Query set for Data objects."""

    # NOTE: This is a static method because it is used from migrations.
    @staticmethod
    def _delete_chunked(queryset, chunk_size=500):
        """Chunked delete, which should be used if deleting many objects.

        The reason why this method is needed is that deleting a lot of Data objects
        requires Django to fetch all of them into memory (fast path is not used) and
        this causes huge memory usage (and possibly OOM).

        :param chunk_size: Optional chunk size
        """
        while True:
            # Discover primary key to limit the current chunk. This is required because delete
            # cannot be called on a sliced queryset due to ordering requirement.
            with transaction.atomic():
                # Get offset of last item (needed because it may be less than the chunk size).
                offset = queryset.order_by('pk')[:chunk_size].count()
                if not offset:
                    break

                # Fetch primary key of last item and use it to delete the chunk.
                last_instance = queryset.order_by('pk')[offset - 1]
                queryset.filter(pk__lte=last_instance.pk).delete()

    def delete_chunked(self, chunk_size=500):
        """Chunked delete, which should be used if deleting many objects.

        The reason why this method is needed is that deleting a lot of Data objects
        requires Django to fetch all of them into memory (fast path is not used) and
        this causes huge memory usage (and possibly OOM).

        :param chunk_size: Optional chunk size
        """
        return DataQuerySet._delete_chunked(self, chunk_size=chunk_size)


class Data(BaseModel):
    """Postgres model for storing data."""

    class Meta(BaseModel.Meta):
        """Data Meta options."""

        permissions = (
            ("view_data", "Can view data"),
            ("edit_data", "Can edit data"),
            ("share_data", "Can share data"),
            ("download_data", "Can download files from data"),
            ("owner_data", "Is owner of the data"),
        )

    #: data object is uploading
    STATUS_UPLOADING = 'UP'
    #: data object is being resolved
    STATUS_RESOLVING = 'RE'
    #: data object is waiting
    STATUS_WAITING = 'WT'
    #: data object is processing
    STATUS_PROCESSING = 'PR'
    #: data object is done
    STATUS_DONE = 'OK'
    #: data object is in error state
    STATUS_ERROR = 'ER'
    #: data object is in dirty state
    STATUS_DIRTY = 'DR'
    STATUS_CHOICES = (
        (STATUS_UPLOADING, 'Uploading'),
        (STATUS_RESOLVING, 'Resolving'),
        (STATUS_WAITING, 'Waiting'),
        (STATUS_PROCESSING, 'Processing'),
        (STATUS_DONE, 'Done'),
        (STATUS_ERROR, 'Error'),
        (STATUS_DIRTY, 'Dirty')
    )

    #: manager
    objects = DataQuerySet.as_manager()

    #: process started date and time (set by
    #: ``resolwe.flow.executors.run.BaseFlowExecutor.run`` or its derivatives)
    started = models.DateTimeField(blank=True, null=True, db_index=True)

    #: process finished date date and time (set by
    #: ``resolwe.flow.executors.run.BaseFlowExecutor.run`` or its derivatives)
    finished = models.DateTimeField(blank=True, null=True, db_index=True)

    #: checksum field calculated on inputs
    checksum = models.CharField(max_length=64, db_index=True, validators=[
        RegexValidator(
            regex=r'^[0-9a-f]{64}$',
            message='Checksum is exactly 40 alphanumerics',
            code='invalid_checksum'
        )
    ])

    status = models.CharField(max_length=2, choices=STATUS_CHOICES, default=STATUS_RESOLVING)
    """
    :class:`Data` status

    It can be one of the following:

    - :attr:`STATUS_UPLOADING`
    - :attr:`STATUS_RESOLVING`
    - :attr:`STATUS_WAITING`
    - :attr:`STATUS_PROCESSING`
    - :attr:`STATUS_DONE`
    - :attr:`STATUS_ERROR`
    """

    #: process used to compute the data object
    process = models.ForeignKey('Process', on_delete=models.PROTECT)

    #: process id
    process_pid = models.PositiveIntegerField(blank=True, null=True)

    #: progress
    process_progress = models.PositiveSmallIntegerField(default=0)

    #: return code
    process_rc = models.PositiveSmallIntegerField(blank=True, null=True)

    #: info log message
    process_info = ArrayField(models.CharField(max_length=255), default=[])

    #: warning log message
    process_warning = ArrayField(models.CharField(max_length=255), default=[])

    #: error log message
    process_error = ArrayField(models.CharField(max_length=255), default=[])

    #: actual inputs used by the process
    input = JSONField(default=dict)

    #: actual outputs of the process
    output = JSONField(default=dict)

    #: total size of data's outputs in bytes
    size = models.BigIntegerField()

    #: data descriptor schema
    descriptor_schema = models.ForeignKey('DescriptorSchema', blank=True, null=True, on_delete=models.PROTECT)

    #: actual descriptor
    descriptor = JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    #: track if user set the data name explicitly
    named_by_user = models.BooleanField(default=False)

    #: dependencies between data objects
    parents = models.ManyToManyField(
        'self',
        through='DataDependency',
        symmetrical=False,
        related_name='children'
    )

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)

    #: actual allocated memory
    process_memory = models.PositiveIntegerField(default=0)

    #: actual allocated cores
    process_cores = models.PositiveSmallIntegerField(default=0)

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)
        self._original_name = self.name

    def save_storage(self, instance, schema):
        """Save basic:json values to a Storage collection."""
        for field_schema, fields in iterate_fields(instance, schema):
            name = field_schema['name']
            value = fields[name]
            if field_schema.get('type', '').startswith('basic:json:'):
                if value and not self.pk:
                    raise ValidationError(
                        'Data object must be `created` before creating `basic:json:` fields')

                if isinstance(value, int):
                    # already in Storage
                    continue

                if isinstance(value, str):
                    file_path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(self.pk), value)
                    if os.path.isfile(file_path):
                        try:
                            with open(file_path) as file_handler:
                                value = json.load(file_handler)
                        except json.JSONDecodeError:
                            with open(file_path) as file_handler:
                                content = file_handler.read()
                                content = content.rstrip()
                                raise ValidationError(
                                    "Value of '{}' must be a valid JSON, current: {}".format(name, content)
                                )

                storage = Storage.objects.create(
                    name='Storage for data id {}'.format(self.pk),
                    contributor=self.contributor,
                    data_id=self.pk,
                    json=value,
                )

                # `value` is copied by value, so `fields[name]` must be changed
                fields[name] = storage.pk

    def resolve_secrets(self):
        """Retrieve handles for all basic:secret: fields on input.

        The process must have the ``secrets`` resource requirement
        specified in order to access any secrets. Otherwise this method
        will raise a ``PermissionDenied`` exception.

        :return: A dictionary of secrets where key is the secret handle
            and value is the secret value.
        """
        secrets = {}
        for field_schema, fields in iterate_fields(self.input, self.process.input_schema):  # pylint: disable=no-member
            if not field_schema.get('type', '').startswith('basic:secret:'):
                continue

            name = field_schema['name']
            value = fields[name]
            try:
                handle = value['handle']
            except KeyError:
                continue

            try:
                secrets[handle] = Secret.objects.get_secret(
                    handle,
                    contributor=self.contributor
                )
            except Secret.DoesNotExist:
                raise PermissionDenied("Access to secret not allowed or secret does not exist")

        # If the process does not not have the right requirements it is not
        # allowed to access any secrets.
        allowed = self.process.requirements.get('resources', {}).get('secrets', False)  # pylint: disable=no-member
        if secrets and not allowed:
            raise PermissionDenied(
                "Process '{}' has secret inputs, but no permission to see secrets".format(
                    self.process.slug  # pylint: disable=no-member
                )
            )

        return secrets

    def save_dependencies(self, instance, schema):
        """Save data: and list:data: references as parents."""
        def add_dependency(value):
            """Add parent Data dependency."""
            try:
                DataDependency.objects.update_or_create(
                    parent=Data.objects.get(pk=value),
                    child=self,
                    defaults={'kind': DataDependency.KIND_IO},
                )
            except Data.DoesNotExist:
                pass

        for field_schema, fields in iterate_fields(instance, schema):
            name = field_schema['name']
            value = fields[name]

            if field_schema.get('type', '').startswith('data:'):
                add_dependency(value)
            elif field_schema.get('type', '').startswith('list:data:'):
                for data in value:
                    add_dependency(data)

    def create_entity(self):
        """Create entity if `flow_collection` is defined in process.

        Following rules applies for adding `Data` object to `Entity`:
        * Only add `Data object` to `Entity` if process has defined
        `flow_collwection` field
        * Add object to existing `Entity`, if all parents that are part
        of it (but not necessary all parents), are part of the same
        `Entity`
        * If parents belong to different `Entities` or do not belong to
        any `Entity`, create new `Entity`

        """
        ds_slug = self.process.flow_collection  # pylint: disable=no-member
        if ds_slug:
            entity_query = Entity.objects.filter(data__in=self.parents.all()).distinct()  # pylint: disable=no-member

            if entity_query.count() == 1:
                entity = entity_query.first()

                copy_permissions(entity, self)
            else:

                descriptor_schema = DescriptorSchema.objects.filter(slug=ds_slug).latest()
                entity = Entity.objects.create(
                    contributor=self.contributor,
                    descriptor_schema=descriptor_schema,
                    name=self.name,
                    tags=self.tags,
                )

                assign_contributor_permissions(entity)

            entity.data.add(self)

    def save(self, render_name=False, *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        """Save the data model."""
        # Generate the descriptor if one is not already set.
        if self.name != self._original_name:
            self.named_by_user = True

        create = self.pk is None
        if create:
            # Default values for INPUT
            input_schema = self.process.input_schema  # pylint: disable=no-member
            for field_schema, fields, path in iterate_schema(self.input, input_schema):
                if 'default' in field_schema and field_schema['name'] not in fields:
                    dict_dot(self.input, path, field_schema['default'])

            if not self.name:
                self._render_name()
            else:
                self.named_by_user = True

            self.checksum = get_data_checksum(
                self.input, self.process.slug, self.process.version)  # pylint: disable=no-member

        elif render_name:
            self._render_name()

        self.save_storage(self.output, self.process.output_schema)  # pylint: disable=no-member

        if self.status != Data.STATUS_ERROR:
            hydrate_size(self)
            # If only specified fields are updated (e.g. in executor), size needs to be added
            if 'update_fields' in kwargs:
                kwargs['update_fields'].append('size')

        # Input Data objects are validated only upon creation as they can be deleted later.
        skip_missing_data = not create
        validate_schema(
            self.input, self.process.input_schema, skip_missing_data=skip_missing_data  # pylint: disable=no-member
        )

        render_descriptor(self)

        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)  # pylint: disable=no-member
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError("`descriptor_schema` must be defined if `descriptor` is given")

        if self.status != Data.STATUS_ERROR:
            path_prefix = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(self.pk))
            output_schema = self.process.output_schema  # pylint: disable=no-member
            if self.status == Data.STATUS_DONE:
                validate_schema(self.output, output_schema, path_prefix=path_prefix)
            else:
                validate_schema(self.output, output_schema, path_prefix=path_prefix,
                                test_required=False)

        with transaction.atomic():
            super().save(*args, **kwargs)

            # We can only save dependencies after the data object has been saved. This
            # is why a transaction block is needed and the save method must be called first.
            if create:
                self.save_dependencies(self.input, self.process.input_schema)  # pylint: disable=no-member

        if create:
            self.create_entity()

    def _render_name(self):
        """Render data name.

        The rendering is based on name template (`process.data_name`) and
        input context.

        """
        if not self.process.data_name or self.named_by_user:  # pylint: disable=no-member
            return

        inputs = copy.deepcopy(self.input)
        hydrate_input_references(inputs, self.process.input_schema, hydrate_values=False)  # pylint: disable=no-member
        template_context = inputs

        try:
            name = render_template(
                self.process,
                self.process.data_name,  # pylint: disable=no-member
                template_context
            )
        except EvaluationError:
            name = '?'

        name_max_len = self._meta.get_field('name').max_length
        if len(name) > name_max_len:
            name = name[:(name_max_len - 3)] + '...'

        self.name = name


class DataDependency(models.Model):
    """Dependency relation between data objects."""

    #: child uses parent's output as its input
    KIND_IO = 'io'
    #: child was spawned by the parent
    KIND_SUBPROCESS = 'subprocess'
    KIND_CHOICES = (
        (KIND_IO, "Input/output dependency"),
        (KIND_SUBPROCESS, "Subprocess"),
    )

    #: child data object
    child = models.ForeignKey(Data, on_delete=models.CASCADE, related_name='parents_dependency')
    #: parent data object
    parent = models.ForeignKey(Data, null=True, on_delete=models.SET_NULL, related_name='children_dependency')
    #: kind of dependency
    kind = models.CharField(max_length=16, choices=KIND_CHOICES)



"""An execution engine that supports workflow specifications."""
import collections

from django.db import transaction

from resolwe.flow.execution_engines.base import BaseExecutionEngine
from resolwe.flow.execution_engines.exceptions import ExecutionError
from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.models import Data, DataDependency, Process
from resolwe.permissions.utils import copy_permissions


class ExecutionEngine(BaseExecutionEngine):
    """An execution engine that supports workflow specifications."""

    name = 'workflow'

    def get_output_schema(self, process):
        """Return any additional output schema for the process."""
        return [
            {'name': 'steps', 'label': "Steps", 'type': 'list:data:'},
        ]

    def _evaluate_expressions(self, expression_engine, step_id, values, context):
        """Recursively evaluate expressions in a dictionary of values."""
        if expression_engine is None:
            return values

        processed = {}
        for name, value in values.items():
            if isinstance(value, str):
                value = value.strip()
                try:
                    expression = expression_engine.get_inline_expression(value)
                    if expression is not None:
                        # Inline expression.
                        value = expression_engine.evaluate_inline(expression, context)
                    else:
                        # Block expression.
                        value = expression_engine.evaluate_block(value, context)
                except EvaluationError as error:
                    raise ExecutionError('Error while evaluating expression for step "{}":\n{}'.format(
                        step_id, error
                    ))
            elif isinstance(value, dict):
                value = self._evaluate_expressions(expression_engine, step_id, value, context)

            processed[name] = value

        return processed

    @transaction.atomic
    def evaluate(self, data):
        """Evaluate the code needed to compute a given Data object."""
        expression_engine = data.process.requirements.get('expression-engine', None)
        if expression_engine is not None:
            expression_engine = self.get_expression_engine(expression_engine)

        # Parse steps.
        steps = data.process.run.get('program', None)
        if steps is None:
            return

        if not isinstance(steps, list):
            raise ExecutionError('Workflow program must be a list of steps.')

        # Expression engine evaluation context.
        context = {
            'input': data.input,
            'steps': collections.OrderedDict(),
        }

        for index, step in enumerate(steps):
            try:
                step_id = step['id']
                step_slug = step['run']
            except KeyError as error:
                raise ExecutionError('Incorrect definition of step "{}", missing property "{}".'.format(
                    step.get('id', index), error
                ))

            # Fetch target process.
            process = Process.objects.filter(slug=step_slug).order_by('-version').first()
            if not process:
                raise ExecutionError('Incorrect definition of step "{}", invalid process "{}".'.format(
                    step_id, step_slug
                ))

            # Process all input variables.
            step_input = step.get('input', {})
            if not isinstance(step_input, dict):
                raise ExecutionError('Incorrect definition of step "{}", input must be a dictionary.'.format(
                    step_id
                ))

            data_input = self._evaluate_expressions(expression_engine, step_id, step_input, context)

            # Create the data object.
            data_object = Data.objects.create(
                process=process,
                contributor=data.contributor,
                input=data_input,
            )
            DataDependency.objects.create(
                parent=data,
                child=data_object,
                kind=DataDependency.KIND_SUBPROCESS,
            )

            # Copy permissions.
            copy_permissions(data, data_object)

            # Copy collections.
            for collection in data.collection_set.all():
                collection.data.add(data_object)

            context['steps'][step_id] = data_object.pk

        # Immediately set our status to done and output all data object identifiers.
        data.output = {
            'steps': list(context['steps'].values()),
        }
        data.status = Data.STATUS_DONE



# pylint: disable=missing-docstring
import os

from django.apps import apps
from django.core.management import call_command

from guardian.shortcuts import assign_perm
from rest_framework import status
from rest_framework.test import force_authenticate

from resolwe.flow.models import Collection, Entity, Relation
from resolwe.flow.models.entity import PositionInRelation, RelationType
from resolwe.flow.views import RelationViewSet
from resolwe.test import ResolweAPITestCase


class TestRelationsAPI(ResolweAPITestCase):

    def setUp(self):
        self.viewset = RelationViewSet
        self.resource_name = 'relation'

        super().setUp()

        # Load fixtures with relation types
        flow_config = apps.get_app_config('flow')
        call_command('loaddata', os.path.join(flow_config.path, 'tests', 'fixtures', 'relationtypes'), verbosity=0)

        self.collection = Collection.objects.create(name='Test collection', contributor=self.contributor)
        collection_2 = Collection.objects.create(name='Second collection', contributor=self.contributor)
        self.entity_1 = Entity.objects.create(name='First entity', contributor=self.contributor)
        self.entity_2 = Entity.objects.create(name='Second entity', contributor=self.contributor)
        self.entity_3 = Entity.objects.create(name='Third entity', contributor=self.contributor)
        self.entity_4 = Entity.objects.create(name='Fourth entity', contributor=self.contributor)

        rel_type_group = RelationType.objects.get(name='group')
        rel_type_series = RelationType.objects.get(name='series')

        self.relation_group = Relation.objects.create(
            contributor=self.contributor,
            collection=self.collection,
            type=rel_type_group,
            label='replicates',
            name='Group relation',
        )
        PositionInRelation.objects.create(relation=self.relation_group, entity=self.entity_1)
        PositionInRelation.objects.create(relation=self.relation_group, entity=self.entity_2)

        self.relation_series = Relation.objects.create(
            contributor=self.contributor,
            collection=collection_2,
            type=rel_type_series,
            label='time-series',
            name='Series relation',
        )
        PositionInRelation.objects.create(relation=self.relation_series, entity=self.entity_1, position='beginning')
        PositionInRelation.objects.create(relation=self.relation_series, entity=self.entity_2, position='beginning')
        PositionInRelation.objects.create(relation=self.relation_series, entity=self.entity_3, position='end')
        PositionInRelation.objects.create(relation=self.relation_series, entity=self.entity_4, position='end')

        assign_perm('view_relation', self.contributor, self.relation_group)
        assign_perm('view_relation', self.contributor, self.relation_series)

        self.add_entity_viewset = RelationViewSet.as_view({
            'post': 'add_entity',
        })

        self.remove_entity_viewset = RelationViewSet.as_view({
            'post': 'remove_entity',
        })

    def test_get(self):
        resp = self._get_detail(self.relation_group.pk, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertEqual(resp.data['entities'], [
            {'entity': self.entity_1.pk, 'position': None},
            {'entity': self.entity_2.pk, 'position': None},
        ])

    def test_filtering(self):
        # Filtering by id
        query_params = {'id': self.relation_group.pk}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 1)
        self.assertEqual(resp.data[0]['id'], self.relation_group.pk)

        # Filtering by collection
        query_params = {'collection': self.collection.pk}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 1)
        self.assertEqual(resp.data[0]['id'], self.relation_group.pk)

        # Filtering by type
        query_params = {'type': 'group'}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 1)
        self.assertEqual(resp.data[0]['id'], self.relation_group.pk)

        # Filtering by label
        query_params = {'label': 'replicates'}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 1)
        self.assertEqual(resp.data[0]['id'], self.relation_group.pk)

        # Filtering by entity
        query_params = {'entity': self.entity_4.pk}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 1)
        self.assertEqual(resp.data[0]['id'], self.relation_series.pk)

        # Filtering by entity and position
        query_params = {'entity': self.entity_4.pk, 'position': 'end'}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 1)
        self.assertEqual(resp.data[0]['id'], self.relation_series.pk)

        # Filtering by entity and position - not matching position
        query_params = {'entity': self.entity_4.pk, 'position': 'beginning'}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(len(resp.data), 0)

        # Filtering by position - missing entity
        query_params = {'position': 'end'}
        resp = self._get_list(user=self.contributor, query_params=query_params)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_without_position(self):
        data = {
            'collection': self.collection.pk,
            'type': 'group',
            'entities': [{'entity': self.entity_3.pk}, {'entity': self.entity_4.pk}],
        }

        resp = self._post(data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Relation.objects.count(), 3)

        relation = Relation.objects.last()
        self.assertEqual(relation.collection, self.collection)
        self.assertEqual(relation.positioninrelation_set.count(), 2)
        self.assertTrue(relation.positioninrelation_set.filter(entity=self.entity_3.pk).exists())
        self.assertTrue(relation.positioninrelation_set.filter(entity=self.entity_4.pk).exists())

    def test_create_with_position(self):
        data = {
            'collection': self.collection.pk,
            'type': 'series',
            'entities': [
                {'entity': self.entity_3.pk, 'position': 1},
                {'entity': self.entity_4.pk, 'position': 2},
            ],
        }
        resp = self._post(data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Relation.objects.count(), 3)

        relation = Relation.objects.last()
        self.assertEqual(relation.collection, self.collection)
        self.assertEqual(relation.positioninrelation_set.count(), 2)
        self.assertTrue(relation.positioninrelation_set.filter(entity=self.entity_3.pk, position=1).exists())
        self.assertTrue(relation.positioninrelation_set.filter(entity=self.entity_4.pk, position=2).exists())

    def test_create_with_missing_type(self):
        data = {
            'collection': self.collection.pk,
            'entities': [
                {'entity': self.entity_3.pk, 'position': 1},
                {'entity': self.entity_4.pk, 'position': 2},
            ],
        }
        resp = self._post(data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_missing_collection(self):
        data = {
            'type': 'series',
            'entities': [
                {'entity': self.entity_3.pk, 'position': 1},
                {'entity': self.entity_4.pk, 'position': 2},
            ],
        }
        resp = self._post(data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

    def test_update(self):
        new_collection = Collection.objects.create(name='Test collection 2', contributor=self.contributor)
        data = {'collection': new_collection.pk}
        resp = self._patch(self.relation_group.pk, data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.relation_group.refresh_from_db()
        self.assertEqual(self.relation_group.collection, new_collection)

        # Relation type cannot be changed
        data = {'type': 'series'}
        resp = self._patch(self.relation_group.pk, data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)

        # Entities nust be updated through special endpoints
        data = {
            'entities': [
                {'entity': self.entity_3.pk, 'position': 1},
            ],
        }
        resp = self._patch(self.relation_group.pk, data, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete(self):
        resp = self._delete(self.relation_group.pk, user=self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Relation.objects.count(), 1)

    def test_add_entity(self):
        data = [{'entity': self.entity_3.pk}]

        request = self.factory.post('', data=data, format='json')
        force_authenticate(request, self.contributor)
        self.add_entity_viewset(request, pk=self.relation_group.pk)

        self.assertEqual(self.relation_group.positioninrelation_set.count(), 3)
        self.assertEqual(self.relation_group.positioninrelation_set.last().position, None)

    def test_add_entity_with_position(self):
        data = [{'entity': self.entity_3.pk, 'position': 'latest'}]

        request = self.factory.post('', data=data, format='json')
        force_authenticate(request, self.contributor)
        self.add_entity_viewset(request, pk=self.relation_group.pk)

        self.assertEqual(self.relation_group.positioninrelation_set.count(), 3)
        self.assertEqual(self.relation_group.positioninrelation_set.last().position, 'latest')

    def test_add_entity_position_none(self):
        data = [{'entity': self.entity_3.pk, 'position': None}]

        request = self.factory.post('', data=data, format='json')
        force_authenticate(request, self.contributor)
        self.add_entity_viewset(request, pk=self.relation_group.pk)

        self.assertEqual(self.relation_group.positioninrelation_set.count(), 3)
        self.assertEqual(self.relation_group.positioninrelation_set.last().position, None)

    def test_add_entity_missing_key(self):
        data = [{'foo': self.entity_3.pk}]
        request = self.factory.post('', data=data, format='json')
        force_authenticate(request, self.contributor)
        self.add_entity_viewset(request, pk=self.relation_group.pk)

        self.assertEqual(self.relation_group.positioninrelation_set.count(), 2)

    def test_remove_entity(self):
        data = {'ids': [self.entity_2.pk]}
        request = self.factory.post('', data=data, format='json')
        force_authenticate(request, self.contributor)
        self.remove_entity_viewset(request, pk=self.relation_group.pk)

        self.assertEqual(self.relation_group.positioninrelation_set.count(), 1)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-06-29 03:00
from __future__ import unicode_literals

from django.db import migrations, models
import resolwe.flow.models.fields
import versionfield


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30)),
                ('slug', resolwe.flow.models.fields.ResolweSlugField()),
                ('version', versionfield.VersionField(default='0.0.0')),
            ],
        ),
    ]



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.flow.executors.local.run.FlowExecutor

"""
import logging
import os
import shlex
import subprocess
import time

from ..run import BaseFlowExecutor

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class FlowExecutor(BaseFlowExecutor):
    """Local dataflow executor proxy."""

    name = 'local'

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)

        self.kill_delay = 5
        self.proc = None
        self.stdout = None
        self.command = '/bin/bash'

    def start(self):
        """Start process execution."""
        self.proc = subprocess.Popen(shlex.split(self.command),
                                     stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT, universal_newlines=True)

        self.stdout = self.proc.stdout

        return self.proc.pid

    def run_script(self, script):
        """Execute the script and save results."""
        self.proc.stdin.write(os.linesep.join(['set -x', 'set +B', script, 'exit']) + os.linesep)
        self.proc.stdin.close()

    def end(self):
        """End process execution."""
        self.proc.wait()

        return self.proc.returncode

    def terminate(self):
        """Terminate a running script."""
        self.proc.terminate()

        time.sleep(self.kill_delay)
        if self.proc.poll() is None:
            self.proc.kill()

        super().terminate()



#!/usr/bin/env python
import channels.log
import logging
import os
import sys

PROJECT_ROOT = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, PROJECT_ROOT)


def get_channels_logger(*args, **kwargs):
    """Return logger for channels."""
    return logging.getLogger("django.channels")


# Force channels to respect logging configurations from settings:
# https://github.com/django/channels/issues/520
channels.log.setup_logger = get_channels_logger

if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")

    # if len(sys.argv) > 1 and sys.argv[1] == 'runserver':
    #     raise ValueError('This Django project is not intended for running a server.')

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



#!/usr/bin/env python
"""Open source dataflow package for Django framework.

See:
https://github.com/genialis/resolwe
"""

from setuptools import find_packages, setup
# Use codecs' open for a consistent encoding
from codecs import open
from os import path

base_dir = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(base_dir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Get package metadata from 'resolwe.__about__.py' file
about = {}
with open(path.join(base_dir, 'resolwe', '__about__.py'), encoding='utf-8') as f:
    exec(f.read(), about)

setup(
    name=about['__title__'],

    version=about['__version__'],

    description=about['__summary__'],
    long_description=long_description,

    url=about['__url__'],

    author=about['__author__'],
    author_email=about['__email__'],

    license=about['__license__'],

    # exclude tests from built/installed package
    packages=find_packages(exclude=['tests', 'tests.*', '*.tests', '*.tests.*']),
    package_data={
        'resolwe': [
            'flow/executors/requirements.txt',
            'flow/static/flow/*.json',
            'toolkit/processes/**.yml',
            'toolkit/tools/**.py',
        ]
    },
    install_requires=[
        'Django~=1.11.0',
        'djangorestframework~=3.7.0',
        'djangorestframework-filters~=0.10.0',
        # XXX: Remove django-autoslug after all migrations that import
        # it are deleted
        'django-autoslug==1.9.3',
        'django-guardian>=1.4.2',
        'django-mathfilters>=0.3.0',
        'django-versionfield2>=0.5.0',
        'django-fernet-fields>=0.5',
        'elasticsearch-dsl~=5.4.0',
        'psycopg2>=2.5.0',
        'mock>=1.3.0',
        'PyYAML>=3.11',
        'jsonschema>=2.4.0',
        'Sphinx>=1.5.1, <1.7.0',
        'Jinja2>=2.9.6',
        'wrapt>=1.10.8',
        'shellescape>=3.4.1',
        # XXX: Temporarily pin idna since the latest version of the requests
        # package (2.18.4) explicitly requires idna>=2.5,<2.7
        'idna==2.6',
        # XXX: djangorestframework-filters has too open requirement for
        # django-filter and doesn't work with the latest version, so we
        # have to pin it
        'django-filter~=1.0.0',
        'redis~=2.10.6',
        'channels~=2.1',
        'channels_redis_persist~=2.2',
        'async-timeout~=2.0',
    ],
    python_requires='>=3.6, <3.7',
    extras_require={
        'docs':  [
            'sphinx_rtd_theme',
        ],
        'package': [
            'twine',
            'wheel',
        ],
        'test': [
            'check-manifest',
            'coverage>=4.2',
            # pycodestyle 2.3.0 raises false-positive for variables
            # starting with 'def'
            # https://github.com/PyCQA/pycodestyle/issues/617
            'pycodestyle~=2.4.0',
            'pydocstyle~=2.1.1',
            'pylint~=1.9.1',
            'readme_renderer',
            'resolwe-runtime-utils>=1.1.0',
            'testfixtures>=4.10.0',
            'tblib>=1.3.0',
            'isort',
        ],
    },

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: OS Independent',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='resolwe dataflow django',
)



# pylint: disable=missing-docstring
from asgiref.sync import async_to_sync

from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.managers import manager
from resolwe.flow.models import Data, DescriptorSchema, Process, Storage
from resolwe.test import TestCase, TransactionTestCase


class ProcessFieldsTagsTest(TransactionTestCase):

    def test_templatetags(self):
        input_process = Process.objects.create(
            name='Input process',
            contributor=self.contributor,
            type='data:test:inputobject:',
            output_schema=[
                {'name': 'test_file', 'type': 'basic:file:'},
            ],
            run={
                'language': 'bash',
                'program': """
mkdir -p path/to
touch path/to/file.txt
re-save-file test_file path/to/file.txt
"""
            }
        )
        descriptor_schema = DescriptorSchema.objects.create(
            name='Test schema',
            slug='test-schema',
            contributor=self.contributor,
            schema=[
                {'name': 'descriptions', 'required': False, 'group': [
                    {'name': 'text', 'type': 'basic:string:'},
                ]}
            ]

        )

        input_data = Data.objects.create(
            name='Input Data object',
            contributor=self.contributor,
            process=input_process,
            descriptor_schema=descriptor_schema,
            descriptor={'descriptions': {'text': 'This is test Data object.'}}
        )

        async_to_sync(manager.communicate)(run_sync=True)

        process = Process.objects.create(
            name='Test template tags',
            requirements={'expression-engine': 'jinja'},
            contributor=self.contributor,
            type='test:data:templatetags:',
            input_schema=[
                {'name': 'input_data', 'type': 'data:test:inputobject:'},
                {'name': 'input_data_list', 'type': 'list:data:test:inputobject:'},
                {'name': 'spacy', 'type': 'basic:string:'},
            ],
            output_schema=[
                {'name': 'name', 'type': 'basic:string:'},
                {'name': 'list_name', 'type': 'list:basic:string:'},
                {'name': 'slug', 'type': 'basic:string:'},
                {'name': 'list_slug', 'type': 'list:basic:string:'},
                {'name': 'id', 'type': 'basic:integer:'},
                {'name': 'list_id', 'type': 'list:basic:string:'},
                {'name': 'type', 'type': 'basic:string:'},
                {'name': 'list_type', 'type': 'list:basic:string:'},
                {'name': 'basename', 'type': 'basic:string:'},
                {'name': 'dirname', 'type': 'basic:string:'},
                {'name': 'subtype', 'type': 'basic:string:'},
                {'name': 'list_subtype', 'type': 'list:basic:string:'},
                {'name': 'yesno', 'type': 'basic:string:'},
                {'name': 'datalookup', 'type': 'basic:integer:'},
                {'name': 'file_url', 'type': 'basic:string:'},
                {'name': 'relative_path', 'type': 'basic:string:'},
                {'name': 'relative_path2', 'type': 'basic:string:'},
                {'name': 'unsafe', 'type': 'basic:string:'},
                {'name': 'safe', 'type': 'basic:string:'},
                {'name': 'description_text', 'type': 'basic:string:'},
                {'name': 'description_full', 'type': 'basic:json:'},
                {'name': 'list_description_text', 'type': 'basic:string:'},
                {'name': 'list_description_full', 'type': 'basic:json:'},
                {'name': 'all', 'type': 'basic:string:'},
                {'name': 'any', 'type': 'basic:string:'},
            ],
            run={
                'language': 'bash',
                'program': """
re-save name {{ input_data | name }}
re-save list_name {{ input_data_list | name }}
re-save slug {{ input_data | slug }}
re-save list_slug {{ input_data_list | slug }}
re-save id {{ input_data | id }}
re-save list_id {{ input_data_list | id }}
re-save type {{ input_data | type }}
re-save list_type {{ input_data_list | type }}
re-save basename {{ '/foo/bar/moo' | basename }}
re-save dirname {{ '/foo/bar/moo' | dirname }}
re-save subtype {{ 'data:test:inputobject:' | subtype('data:') }}
re-save list_subtype {{ ['data:test:inputobject:'] | subtype('data:') }}
re-save yesno {{ true | yesno('yes', 'no') }}
re-save datalookup {{ 'input-data-object' | data_by_slug }}
re-save file_url {{ input_data.test_file | get_url }}
re-save relative_path {{ input_data.test_file | relative_path }}
re-save relative_path2 {{ input_data.test_file.file | relative_path }}
re-save unsafe {{ spacy }}
re-save description_text {{ input_data | descriptor('descriptions.text') }}
re-save description_full {{ input_data | descriptor }}
re-save list_description_text {{ input_data_list[0] | descriptor('descriptions.text') }}
re-save list_description_full {{ input_data_list[0] | descriptor }}
re-save all {{ [true, false] | all }}
re-save any {{ [true, false] | any }}

function save-safe() {
    re-save safe $1
}
save-safe {{ spacy | safe }}
"""
            }

        )
        data = Data.objects.create(
            name='Data object',
            contributor=self.contributor,
            process=process,
            input={
                'input_data': input_data.pk,
                'input_data_list': [input_data.pk],
                'spacy': 'this has \'some\' spaces',
            },
        )

        async_to_sync(manager.communicate)(run_sync=True)

        data.refresh_from_db()

        self.assertEqual(data.output['name'], input_data.name)
        self.assertEqual(data.output['slug'], input_data.slug)
        self.assertEqual(data.output['id'], input_data.pk)
        self.assertEqual(data.output['list_id'], [input_data.id])
        self.assertEqual(data.output['type'], input_process.type)
        self.assertEqual(data.output['basename'], 'moo')
        self.assertEqual(data.output['dirname'], '/foo/bar')
        self.assertEqual(data.output['subtype'], 'True')
        self.assertEqual(data.output['yesno'], 'yes')
        self.assertEqual(data.output['datalookup'], input_data.pk)
        self.assertEqual(data.output['file_url'], 'localhost/data/{}/path/to/file.txt'.format(input_data.pk))
        self.assertEqual(data.output['relative_path'], 'path/to/file.txt')
        self.assertEqual(data.output['relative_path2'], 'path/to/file.txt')
        self.assertEqual(data.output['unsafe'], 'this has \'some\' spaces')
        self.assertEqual(data.output['safe'], 'this')
        self.assertEqual(data.output['description_text'], 'This is test Data object.')
        self.assertEqual(data.output['list_description_text'], 'This is test Data object.')

        storage = Storage.objects.get(pk=data.output['description_full'])
        self.assertEqual(storage.json, {'descriptions': {'text': 'This is test Data object.'}})

        storage = Storage.objects.get(pk=data.output['list_description_full'])
        self.assertEqual(storage.json, {'descriptions': {'text': 'This is test Data object.'}})

        # Return values of jinja filters are casted to strings when inserted to bash
        self.assertEqual(data.output['list_name'], str([input_data.name]))
        self.assertEqual(data.output['list_slug'], str([input_data.slug]))
        self.assertEqual(data.output['list_type'], str([input_process.type]))
        self.assertEqual(data.output['list_subtype'], '[True]')

        self.assertEqual(data.output['all'], 'False')
        self.assertEqual(data.output['any'], 'True')


class ExpressionEngineTest(TestCase):

    def test_jinja_engine(self):
        engine = manager.get_expression_engine('jinja')
        block = engine.evaluate_block('Hello {{ world }}', {'world': 'cruel world'})
        self.assertEqual(block, 'Hello cruel world')
        block = engine.evaluate_block('Hello {% if world %}world{% endif %}', {'world': True})
        self.assertEqual(block, 'Hello world')

        with self.assertRaises(EvaluationError):
            engine.evaluate_block('Hello {% bar')

        expression = engine.evaluate_inline('world', {'world': 'cruel world'})
        self.assertEqual(expression, 'cruel world')
        expression = engine.evaluate_inline('world', {'world': True})
        self.assertEqual(expression, True)
        expression = engine.evaluate_inline('world | yesno("yes", "no")', {'world': False})
        self.assertEqual(expression, 'no')
        expression = engine.evaluate_inline('[1, 2, 3, world]', {'world': 4})
        self.assertEqual(expression, [1, 2, 3, 4])
        expression = engine.evaluate_inline('a.b.c.d', {})
        self.assertEqual(expression, None)
        expression = engine.evaluate_inline('a.b.c().d', {})
        self.assertEqual(expression, None)
        expression = engine.evaluate_inline('a.b.0.d', {'a': {'b': [{'d': 'Hello world'}]}})
        self.assertEqual(expression, 'Hello world')
        expression = engine.evaluate_inline('a.b.0.d', {})
        self.assertEqual(expression, None)
        expression = engine.evaluate_inline('[world, false] | all', {'world': True})
        self.assertFalse(expression)
        expression = engine.evaluate_inline('[world, false] | any', {'world': True})
        self.assertTrue(expression)

        # Test that propagation of undefined values works.
        expression = engine.evaluate_inline('foo.bar | name | default("bar")', {})
        self.assertEqual(expression, 'bar')
        expression = engine.evaluate_inline('foo.bar | id | default("bar")', {})
        self.assertEqual(expression, 'bar')
        expression = engine.evaluate_inline('foo.bar | type | default("bar")', {})
        self.assertEqual(expression, 'bar')
        expression = engine.evaluate_inline('foo.bar | basename | default("bar")', {})
        self.assertEqual(expression, 'bar')

        # Ensure that filter decorations are correctly copied when decorating filters to
        # automatically propagate undefined values on exceptions.
        expression = engine.evaluate_inline('foo | join(" ")', {'foo': ['a', 'b', 'c']})
        self.assertEqual(expression, 'a b c')



# pylint: disable=missing-docstring
import os
from io import StringIO

import yaml

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase as DjangoTestCase

from guardian.models import UserObjectPermission
from guardian.shortcuts import assign_perm, get_perms

from resolwe.flow.models import Process
from resolwe.test import ProcessTestCase, TestCase

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')


class ProcessRegisterTest(TestCase):

    def test_process_register_all(self):
        out, err = StringIO(), StringIO()
        call_command('register', path=[PROCESSES_DIR], stdout=out, stderr=err)
        self.assertTrue('Inserted test-min' in out.getvalue())
        self.assertTrue('Skip processor test-min: newer version installed' in err.getvalue())
        self.assertTrue(
            'Skip processor test-broken-invalid-execution-engine: '
            'execution engine \'invalid\' not supported' in err.getvalue()
        )

        # Check that contributor gets all permissions.
        process = Process.objects.first()
        self.assertEqual(len(get_perms(self.admin, process)), 3)

        out, err = StringIO(), StringIO()
        call_command('register', path=[PROCESSES_DIR], stdout=out, stderr=err)
        self.assertTrue('Skip processor test-min: same version installed' in out.getvalue())
        self.assertTrue('Skip processor test-bloated: same version installed' in out.getvalue())
        self.assertTrue('Skip processor test-min: newer version installed' in err.getvalue())

        out, err = StringIO(), StringIO()
        call_command('register', path=[PROCESSES_DIR], force=True, stdout=out, stderr=err)
        self.assertTrue('Updated test-min' in out.getvalue())
        self.assertTrue('Updated test-bloated' in out.getvalue())
        self.assertTrue('Skip processor test-min: newer version installed' in err.getvalue())

    def test_validation_of_defaults(self):
        process_path = os.path.join(PROCESSES_DIR, 'wrong_defaults')

        out, err = StringIO(), StringIO()
        call_command('register', path=[process_path], schemas=['test-wrong-type'], stdout=out, stderr=err)
        self.assertIn('VALIDATION ERROR:', err.getvalue())

        out, err = StringIO(), StringIO()
        call_command('register', path=[process_path], schemas=['test-out-of-range'], stdout=out, stderr=err)
        self.assertIn('VALIDATION ERROR:', err.getvalue())

    def test_process_register_filter(self):
        out, err = StringIO(), StringIO()
        # Fields types are also tested here, as process won't register
        # if any of them fail.
        call_command('register', path=[PROCESSES_DIR], schemas=['test-bloated'], stdout=out, stderr=err)
        self.assertTrue('Inserted test-bloated' in out.getvalue())
        self.assertTrue('Inserted test-min' not in out.getvalue())
        self.assertEqual('', err.getvalue())

        out, err = StringIO(), StringIO()
        call_command('register', path=[PROCESSES_DIR], schemas=['test-bloated'], stdout=out, stderr=err)
        self.assertTrue('Skip processor test-bloated: same version installed' in out.getvalue())
        self.assertEqual('', err.getvalue())

        out, err = StringIO(), StringIO()
        call_command(
            'register', path=[PROCESSES_DIR], schemas=['test-bloated'], force=True, stdout=out, stderr=err)
        self.assertTrue('Updated test-bloated' in out.getvalue())
        self.assertEqual('', err.getvalue())

    def test_inherit_perms(self):
        out, err = StringIO(), StringIO()
        first_version_path = os.path.join(PROCESSES_DIR, 'first_version')
        call_command('register', path=[first_version_path], stdout=out, stderr=err)

        process = Process.objects.latest()
        assign_perm('view_process', self.user, process)

        out, err = StringIO(), StringIO()
        second_version_path = os.path.join(PROCESSES_DIR, 'second_version')
        call_command('register', path=[second_version_path], stdout=out, stderr=err)

        process = Process.objects.latest()

        self.assertEqual(UserObjectPermission.objects.filter(user=self.user).count(), 2)
        self.assertTrue(self.user.has_perm('flow.view_process', process))


class ProcessRegisterTestNoAdmin(DjangoTestCase):

    def test_process_register_no_admin(self):
        err = StringIO()
        self.assertRaises(SystemExit, call_command, 'register', path=[PROCESSES_DIR], stderr=err)
        self.assertEqual('Admin does not exist: create a superuser\n', err.getvalue())


class ListDockerImagesTest(ProcessTestCase):

    def setUp(self):
        super().setUp()

        # Make sure the test processes are in the database
        self._register_schemas(path=[PROCESSES_DIR])

    def test_basic_list(self):
        # List Docker images and check if there's at least one
        out, err = StringIO(), StringIO()
        call_command('list_docker_images', stdout=out, stderr=err)
        self.assertNotEqual('', out.getvalue())
        self.assertEqual('', err.getvalue())

    def test_basic_list_yaml(self):
        # List Docker images in YAML format and see if the output is valid YAML
        out, err = StringIO(), StringIO()
        call_command('list_docker_images', format='yaml', stdout=out, stderr=err)
        self.assertNotEqual('', out.getvalue())
        self.assertEqual('', err.getvalue())
        imgs = yaml.safe_load(out.getvalue())
        self.assertTrue(isinstance(imgs, list))
        self.assertTrue(len(imgs) != 0)
        self.assertTrue(isinstance(imgs[0], dict))

    def test_invalid_format(self):
        # An unsupported format option should return an error
        self.assertRaises(CommandError, call_command, 'list_docker_images', format='invalid')

    def test_list_has_test_image(self):
        # The returned list must contain the image 'resolwe/test:base'
        out, err = StringIO(), StringIO()
        call_command('list_docker_images', stdout=out, stderr=err)
        self.assertEqual('', err.getvalue())
        self.assertTrue('resolwe/test:base' in out.getvalue())

    def test_list_has_test_image_yaml(self):
        # The returned list must contain the image 'resolwe/test:base',
        # in the YAML output as well
        out, err = StringIO(), StringIO()
        call_command('list_docker_images', format='yaml', stdout=out, stderr=err)
        self.assertEqual('', err.getvalue())
        imgs = yaml.safe_load(out.getvalue())
        self.assertTrue(isinstance(imgs, list))
        self.assertTrue(len(imgs) != 0)
        self.assertTrue(isinstance(imgs[0], dict))
        self.assertTrue(dict(name='resolwe/test', tag='base') in imgs)

    def test_list_is_sorted(self):
        # The returned list must be sorted alphabetically
        out, err = StringIO(), StringIO()
        call_command('list_docker_images', stdout=out, stderr=err)
        self.assertEqual('', err.getvalue())
        self.assertNotEqual('', out.getvalue())
        lines = out.getvalue().split('\n')[:-1]
        self.assertEqual(lines, sorted(lines))

    def test_list_versions(self):
        # The returned list must contain only the Docker image from
        # the latest version of a given process
        out, err = StringIO(), StringIO()
        call_command('list_docker_images', stdout=out, stderr=err)
        self.assertEqual('', err.getvalue())
        self.assertIn('resolwe/test:versioning-2', out.getvalue())
        self.assertNotIn('resolwe/test:versioning-1', out.getvalue())



""".. Ignore pydocstyle D400.

===============
Slurm Connector
===============

"""
import logging
import os
import shlex
import subprocess

from django.conf import settings

from resolwe.utils import BraceMessage as __

from .base import BaseConnector

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# We add this much to the memory limit to account for executor overhead,
# since the executor is running in the same environment as the process.
EXECUTOR_MEMORY_OVERHEAD = 200


class Connector(BaseConnector):
    """Slurm-based connector for job execution."""

    def submit(self, data, runtime_dir, argv):
        """Run process with SLURM.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        limits = data.process.get_resource_limits()
        logger.debug(__(
            "Connector '{}' running for Data with id {} ({}).",
            self.__class__.__module__,
            data.id,
            repr(argv)
        ))

        # Compute target partition.
        partition = getattr(settings, 'FLOW_SLURM_PARTITION_DEFAULT', None)
        if data.process.slug in getattr(settings, 'FLOW_SLURM_PARTITION_OVERRIDES', {}):
            partition = settings.FLOW_SLURM_PARTITION_OVERRIDES[data.process.slug]

        try:
            # Make sure the resulting file is executable on creation.
            script_path = os.path.join(runtime_dir, 'slurm.sh')
            file_descriptor = os.open(script_path, os.O_WRONLY | os.O_CREAT, mode=0o555)
            with os.fdopen(file_descriptor, 'wt') as script:
                script.write('#!/bin/bash\n')
                script.write('#SBATCH --mem={}M\n'.format(limits['memory'] + EXECUTOR_MEMORY_OVERHEAD))
                script.write('#SBATCH --cpus-per-task={}\n'.format(limits['cores']))
                if partition:
                    script.write('#SBATCH --partition={}\n'.format(partition))

                # Render the argument vector into a command line.
                line = ' '.join(map(shlex.quote, argv))
                script.write(line + '\n')

            command = ['/usr/bin/env', 'sbatch', script_path]
            subprocess.Popen(
                command,
                cwd=runtime_dir,
                stdin=subprocess.DEVNULL
            ).wait()
        except OSError as err:
            logger.error(__(
                "OSError occurred while preparing SLURM script for Data {}: {}",
                data.id, err
            ))



""".. Ignore pydocstyle D400.

==============
Flow Utilities
==============

.. automodule:: resolwe.flow.utils.purge
   :members:

.. automodule:: resolwe.flow.utils.exceptions
   :members:

.. automodule:: resolwe.flow.utils.stats
   :members:

"""
import functools
import hashlib
import json
import os

from django.apps import apps
from django.conf import settings
from django.db import models

from .iterators import iterate_dict, iterate_fields, iterate_schema  # pylint: disable=unused-import


def get_data_checksum(proc_input, proc_slug, proc_version):
    """Compute checksum of processor inputs, name and version."""
    checksum = hashlib.sha256()
    checksum.update(json.dumps(proc_input, sort_keys=True).encode('utf-8'))
    checksum.update(proc_slug.encode('utf-8'))
    checksum.update(str(proc_version).encode('utf-8'))
    return checksum.hexdigest()


def dict_dot(d, k, val=None, default=None):
    """Get or set value using a dot-notation key in a multilevel dict."""
    if val is None and k == '':
        return d

    def set_default(dict_or_model, key, default_value):
        """Set default field value."""
        if isinstance(dict_or_model, models.Model):
            if not hasattr(dict_or_model, key):
                setattr(dict_or_model, key, default_value)

            return getattr(dict_or_model, key)
        else:
            return dict_or_model.setdefault(key, default_value)

    def get_item(dict_or_model, key):
        """Get field value."""
        if isinstance(dict_or_model, models.Model):
            return getattr(dict_or_model, key)
        else:
            return dict_or_model[key]

    def set_item(dict_or_model, key, value):
        """Set field value."""
        if isinstance(dict_or_model, models.Model):
            setattr(dict_or_model, key, value)
        else:
            dict_or_model[key] = value

    if val is None and callable(default):
        # Get value, default for missing
        return functools.reduce(lambda a, b: set_default(a, b, default()), k.split('.'), d)

    elif val is None:
        # Get value, error on missing
        return functools.reduce(get_item, k.split('.'), d)

    else:
        # Set value
        try:
            k, k_last = k.rsplit('.', 1)
            set_item(dict_dot(d, k, default=dict), k_last, val)
        except ValueError:
            set_item(d, k, val)
        return val


def get_apps_tools():
    """Get applications' tools and their paths.

    Return a dict with application names as keys and paths to tools'
    directories as values. Applications without tools are omitted.
    """
    tools_paths = {}

    for app_config in apps.get_app_configs():
        proc_path = os.path.join(app_config.path, 'tools')
        if os.path.isdir(proc_path):
            tools_paths[app_config.name] = proc_path

    custom_tools_paths = getattr(settings, 'RESOLWE_CUSTOM_TOOLS_PATHS', [])
    if not isinstance(custom_tools_paths, list):
        raise KeyError("`RESOLWE_CUSTOM_TOOLS_PATHS` setting must be a list.")

    for seq, custom_path in enumerate(custom_tools_paths):
        custom_key = '_custom_{}'.format(seq)
        tools_paths[custom_key] = custom_path

    return tools_paths



""".. Ignore pydocstyle D400.

============
Flow Filters
============

"""
import rest_framework_filters as filters
from django_filters.filters import BaseCSVFilter

from .models import Collection, Data, DescriptorSchema, Entity, Process


class BaseResolweFilter(filters.FilterSet):
    """Base filter for Resolwe's endpoints."""

    id = filters.AllLookupsFilter()  # pylint: disable=invalid-name
    slug = filters.AllLookupsFilter()
    name = filters.AllLookupsFilter()
    contributor = filters.NumberFilter()
    created = filters.AllLookupsFilter()
    modified = filters.AllLookupsFilter()

    class Meta:
        """Filter configuration."""

        fields = ['id', 'slug', 'name', 'contributor', 'created', 'modified']


class DescriptorSchemaFilter(BaseResolweFilter):
    """Filter the DescriptorSchema endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = DescriptorSchema


class CollectionFilter(BaseResolweFilter):
    """Filter the Collection endpoint."""

    data = filters.ModelChoiceFilter(queryset=Data.objects.all())
    entity = filters.ModelChoiceFilter(queryset=Entity.objects.all())
    descriptor_schema = filters.RelatedFilter(
        DescriptorSchemaFilter, name='descriptor_schema', queryset=DescriptorSchema.objects.all()
    )
    description = filters.AllLookupsFilter()

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Collection
        fields = BaseResolweFilter.Meta.fields + ['data', 'entity', 'descriptor_schema', 'description']


class TagsFilter(BaseCSVFilter, filters.CharFilter):
    """Filter for tags."""

    def __init__(self, *args, **kwargs):
        """Construct tags filter."""
        kwargs.setdefault('lookup_expr', 'contains')
        super().__init__(*args, **kwargs)


class EntityFilter(CollectionFilter):
    """Filter the Entity endpoint."""

    collections = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    descriptor_completed = filters.BooleanFilter()
    tags = TagsFilter()

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Entity
        fields = BaseResolweFilter.Meta.fields + ['collections', 'descriptor_completed', 'tags']


class ProcessFilter(BaseResolweFilter):
    """Filter the Process endpoint."""

    category = filters.CharFilter(name='category', lookup_expr='startswith')

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Process
        fields = BaseResolweFilter.Meta.fields + ['category']


class DataFilter(BaseResolweFilter):
    """Filter the Data endpoint."""

    collection = filters.RelatedFilter(CollectionFilter, queryset=Collection.objects.all())
    entity = filters.ModelChoiceFilter(queryset=Entity.objects.all())
    type = filters.CharFilter(name='process__type', lookup_expr='startswith')
    status = filters.CharFilter(lookup_expr='iexact')
    finished = filters.AllLookupsFilter()
    started = filters.AllLookupsFilter()
    process = filters.RelatedFilter(ProcessFilter, queryset=Process.objects.all())
    tags = TagsFilter()
    parents = filters.ModelChoiceFilter(queryset=Data.objects.all())
    children = filters.ModelChoiceFilter(queryset=Data.objects.all())

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Data
        fields = BaseResolweFilter.Meta.fields + [
            'collection', 'entity', 'type', 'status', 'finished', 'started', 'process', 'tags',
            'parents', 'children'
        ]



""".. Ignore pydocstyle D400.

=======================
Interpret Test Profiles
=======================

"""
import json

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """Utility for interpreting test profiles."""

    help = "Utility for interpreting test profiles."

    def add_arguments(self, parser):
        """Add command arguments."""
        parser.add_argument('log', nargs='+', type=str,
                            help="profile log filename")
        parser.add_argument('--sort', type=str, choices=('name', 'time'), default='name',
                            help="sort order (default: name)")

    def handle(self, log, **options):
        """Handle command."""
        results = {}
        for filename in log:
            with open(filename) as log_file:
                for line in log_file:
                    data = json.loads(line)
                    test = results.setdefault(data['test'], {})
                    del data['test']
                    test.update(data)

        # Generate sort function.
        if options['sort'] == 'name':
            def sort_key(item):
                """Sort by block name."""
                key, _ = item
                return key

            sort_reverse = False
        elif options['sort'] == 'time':
            def sort_key(item):
                """Sort by block total time."""
                _, blocks = item
                return blocks['total']

            sort_reverse = True

        totals = {}
        for test, blocks in sorted(results.items(), key=sort_key, reverse=sort_reverse):
            print('{}'.format(test))

            total_time = blocks['total']
            for block, time in blocks.items():
                totals[block] = totals.get(block, 0.0) + time
                print('  {}: {:.2f} ({:.2f}%)'.format(block, time, time * 100.0 / total_time))

            print('')

        print('Overall:')
        total_time = totals['total']
        for block, time in totals.items():
            print('  {}: {:.2f} ({:.2f}%)'.format(block, time, time * 100.0 / total_time))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-01 13:46
from __future__ import unicode_literals

import autoslug.fields
from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import versionfield


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('flow', '0023_update_checksum'),
    ]

    operations = [
        migrations.CreateModel(
            name='PositionInRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Relation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', autoslug.fields.AutoSlugField(editable=True, max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default='0.0.0')),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('label', models.CharField(max_length=100, null=True, blank=True)),
            ],
            options={
                'abstract': False,
                'get_latest_by': 'version',
                'default_permissions': (),
                'permissions': (('view_relation', 'Can view relation'), ('edit_relation', 'Can edit relation'), ('share_relation', 'Can share relation'), ('owner_relation', 'Is owner of the relation')),
            },
        ),
        migrations.CreateModel(
            name='RelationType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
                ('ordered', models.BooleanField(default=False)),
            ],
        ),
        migrations.AddField(
            model_name='relation',
            name='collection',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='flow.Collection'),
        ),
        migrations.AddField(
            model_name='relation',
            name='contributor',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='relation',
            name='entities',
            field=models.ManyToManyField(through='flow.PositionInRelation', to='flow.Entity'),
        ),
        migrations.AddField(
            model_name='relation',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='flow.RelationType'),
        ),
        migrations.AddField(
            model_name='positioninrelation',
            name='entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Entity'),
        ),
        migrations.AddField(
            model_name='positioninrelation',
            name='relation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Relation'),
        ),
        migrations.AlterUniqueTogether(
            name='relation',
            unique_together=set([('slug', 'version')]),
        ),
    ]



# pylint: disable=missing-docstring
import builtins
import gzip
import io
import json
import os.path
import tempfile
import unittest.mock as mock

from resolwe.flow.models import Storage
from resolwe.test import ProcessTestCase, TestCase, is_testing


class TestingFrameworkTestCase(TestCase):

    def setUp(self):
        self.dummy_case = ProcessTestCase.__new__(ProcessTestCase)
        self.dummy_case.files_path = ''
        self.dummy_case._debug_info = lambda _: ''  # pylint: disable=protected-access
        setattr(self.dummy_case, 'assertEqual', self.assertEqual)

    @mock.patch("os.path.isfile")
    def test_assert_file_date_nofilter(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"some line\ndate: 2016-02-10\n")
        output2_file = io.BytesIO(b"some line\ndate: 2015-10-31\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, 'open', open_mock):
            obj_mock = mock.MagicMock()
            self.assertRaises(
                AssertionError,
                ProcessTestCase._assert_file,  # pylint: disable=protected-access
                self.dummy_case, obj_mock, "", "",
            )

    @mock.patch("os.path.isfile")
    def test_assert_file_date_filter(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"some line\ndate: 2016-02-10\n")
        output2_file = io.BytesIO(b"some line\ndate: 2015-10-31\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, 'open', open_mock):
            obj_mock = mock.MagicMock()

            def date_in_line(line):
                return line.startswith(b"date")

            ProcessTestCase._assert_file(  # pylint: disable=protected-access
                self.dummy_case, obj_mock, "", "", file_filter=date_in_line)

    @mock.patch("os.path.isfile")
    def test_assert_file_sort_false(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"A\nB\nC\n")
        output2_file = io.BytesIO(b"A\nC\nB\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, 'open', open_mock):
            obj_mock = mock.MagicMock()
            self.assertRaises(
                AssertionError,
                ProcessTestCase._assert_file,  # pylint: disable=protected-access
                self.dummy_case, obj_mock, "", "",
            )

    @mock.patch("os.path.isfile")
    def test_assert_file_sort_true(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"A\nB\nC\n")
        output2_file = io.BytesIO(b"A\nC\nB\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, 'open', open_mock):
            obj_mock = mock.MagicMock()
            ProcessTestCase._assert_file(  # pylint: disable=protected-access
                self.dummy_case, obj_mock, "", "", sort=True)

    def test_assert_json_storage_object(self):
        example_json = {'foo': [1.0, 2.5, 3.14], 'bar': ['ba', 'cd']}

        obj_mock = mock.MagicMock()

        storage_mock = mock.MagicMock(spec=Storage)
        storage_mock.id = 'no_id'
        storage_mock.json = example_json

        # use in-memory binary stream object for speed and simplicity
        gzipped_json_file = io.BytesIO()
        with gzip.open(gzipped_json_file, mode='wt') as f:
            json.dump(example_json, f)
        # set seek position of the binary stream object back to 0
        gzipped_json_file.seek(0)

        join_mock = mock.MagicMock(side_effect=[gzipped_json_file])
        isfile_mock = mock.MagicMock(return_value=True)
        with mock.patch.object(os.path, 'join', join_mock):
            with mock.patch.object(os.path, 'isfile', isfile_mock):
                ProcessTestCase.assertJSON(self.dummy_case, obj_mock, storage_mock, '', 'foo.gz')

    @mock.patch('resolwe.flow.models.Storage.objects.get')
    def test_assert_json_storage_id(self, get_mock):
        example_json = {'foo': [1.0, 2.5, 3.14], 'bar': ['ba', 'cd']}

        obj_mock = mock.MagicMock()

        storage_id = 'no_id'
        storage_mock = mock.MagicMock(spec=Storage)
        storage_mock.id = storage_id
        storage_mock.json = example_json
        get_mock.side_effect = [storage_mock]

        # use in-memory binary stream object for speed and simplicity
        gzipped_json_file = io.BytesIO()
        with gzip.open(gzipped_json_file, mode='wt') as f:
            json.dump(example_json, f)
        # set seek position of the binary stream object back to 0
        gzipped_json_file.seek(0)

        join_mock = mock.MagicMock(side_effect=[gzipped_json_file])
        isfile_mock = mock.MagicMock(return_value=True)
        with mock.patch.object(os.path, 'join', join_mock):
            with mock.patch.object(os.path, 'isfile', isfile_mock):
                ProcessTestCase.assertJSON(self.dummy_case, obj_mock, storage_id, '', 'foo.gz')

    def test_assert_json_file_missing(self):
        example_json = {'foo': [1.0, 2.5, 3.14], 'bar': ['ba', 'cd']}

        obj_mock = mock.MagicMock()

        storage_mock = mock.MagicMock(spec=Storage)
        storage_mock.id = 'no_id'
        storage_mock.json = example_json

        # use in-memory binary stream object for speed and simplicity
        gzipped_json_file = io.BytesIO()
        join_modified_values = [gzipped_json_file]
        orig_join = os.path.join

        # NOTE: coverage tool needs original 'os.path.join'
        def join_side_effect(path, *paths):
            if join_modified_values:
                return join_modified_values.pop(0)
            else:
                return orig_join(path, *paths)

        join_mock = mock.MagicMock(side_effect=join_side_effect)
        isfile_mock = mock.MagicMock(return_value=False)
        with mock.patch.object(os.path, 'join', join_mock):
            with mock.patch.object(os.path, 'isfile', isfile_mock):

                # https://github.com/PyCQA/pylint/issues/1653
                self.assertRaisesRegex(  # pylint: disable=deprecated-method
                    AssertionError,
                    'Output file .* missing so it was created.',
                    ProcessTestCase.assertJSON,
                    self.dummy_case, obj_mock, storage_mock, '', 'foo.gz',
                )

        # set seek position of the binary stream object back to 0
        gzipped_json_file.seek(0)
        with gzip.open(gzipped_json_file, mode='rt') as f:
            unzipped_json = json.load(f)
        self.assertEqual(example_json, unzipped_json)

    def test_debug_info_non_ascii(self):
        non_ascii_text = 'Some non-ascii chars č ü €'

        dummy_case = ProcessTestCase.__new__(ProcessTestCase)

        obj_mock = mock.MagicMock()
        obj_mock.pk = 'no_id'

        with tempfile.NamedTemporaryFile(mode='wb') as stdout_file:
            stdout_file.write(non_ascii_text.encode('utf-8'))
            # set seek position of the file back to 0
            stdout_file.seek(0)
            with mock.patch.object(os.path, 'join', side_effect=[stdout_file.name]):
                with mock.patch.object(os.path, 'isfile', return_value=True):

                    # https://github.com/PyCQA/pylint/issues/1653
                    self.assertRegex(  # pylint: disable=deprecated-method
                        ProcessTestCase._debug_info(dummy_case, obj_mock),   # pylint: disable=protected-access
                        non_ascii_text,
                    )

    def test_assert_almost_equal_generic(self):  # pylint: disable=invalid-name
        self.assertAlmostEqualGeneric(1.00000001, 1.0)
        self.assertAlmostEqualGeneric([1.00000001], [1.0])
        self.assertAlmostEqualGeneric({'foo': 1.00000001}, {'foo': 1.0})
        self.assertAlmostEqualGeneric({'foo': [1.00000001]}, {'foo': [1.0]})
        self.assertAlmostEqualGeneric({'foo': 1.00000001, 'bar': 'moo'}, {'foo': 1.0, 'bar': 'moo'})
        self.assertAlmostEqualGeneric(1.00000001, 1.0, msg="Test message")

    def test_is_testing(self):
        self.assertTrue(is_testing())



"""Urls for Django Rest Framework."""
from django.conf.urls import include, url

from rest_framework import routers

from .flow.views import (
    CollectionViewSet, DataViewSet, DescriptorSchemaViewSet, EntityViewSet, ProcessViewSet, RelationViewSet,
    StorageViewSet,
)

api_router = routers.DefaultRouter(trailing_slash=False)  # pylint: disable=invalid-name
api_router.register(r'collection', CollectionViewSet)
api_router.register(r'process', ProcessViewSet)
api_router.register(r'data', DataViewSet)
api_router.register(r'entity', EntityViewSet)
api_router.register(r'relation', RelationViewSet)
api_router.register(r'descriptorschema', DescriptorSchemaViewSet)
api_router.register(r'storage', StorageViewSet)


urlpatterns = [  # pylint: disable=invalid-name
    url(r'^', include(api_router.urls, namespace='resolwe-api')),
]
